/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_FAST_INTERSECTOR_HXX_
# define WIFI_FRAME_FILTER_FAST_INTERSECTOR_HXX_

# include <iostream>

# include "fast_intersector.hh"

# include <trace-tools/wifi/frame/filter/non_noisy_prism.hh>
# include <trace-tools/wifi/frame/filter/uniquely_identifiable.hh>

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      namespace internals
      {

	template <class I1, class I2, class B1, class B2>
	fast_intersector_iterator<I1, I2, B1, B2>::
	fast_intersector_iterator(const iterable_type& i, bool end):
	  super_type (),
	  iterable_ (&i),
	  next_ (end ? i.last_ : i.first_)
	{
	  if (not end)
	    increment();
	}

	template <class I1, class I2, class B1, class B2>
	bool
	fast_intersector_iterator<I1, I2, B1, B2>::
	equal(const fast_intersector_iterator& rhs) const
	{
	  if (not rhs.value())
	    return not this->value();
	  return next_ == rhs.next_;
	}

	template <class I1, class I2, class B1, class B2>
	void
	fast_intersector_iterator<I1, I2, B1, B2>::increment()
	{
	  const I1&		last = this->iterable_->last_;
	  bool			found (false);

	  while (next_ != last and not found)
	    {
	      found = check();
	      ++next_;
	    }

	  if (not found)
	    this->value() = boost::none_t ();
	}

	template <class I1, class I2, class B1, class B2>
	bool
	fast_intersector_iterator<I1, I2, B1, B2>::check()
	{
	  typedef typename iterable_type::arrivals_type		arrivals_type;
	  typedef typename arrivals_type::const_iterator	iterator;

	  const arrivals_type&	arrivals = iterable_->arrivals_;
	  const iterator	i = arrivals.find(*next_);

	  if (i == arrivals.end())
	    return false;

	  if (i->microseconds() < last_)
	    {
	      std::cerr << "WARNING: Unique frame " << i->frame_id()
			<< " (" << i->microseconds() << " us) anterior to "
		           "previous intersection (" << last_ << " us), "
		           "ignoring.\n"
		           "WARNING: Probably this is due to previously "
			   "ignored frames." << std::endl;
	      return false;
	    }

	  last_ = i->microseconds();
	  this->value() = value_type (*next_, *i);
	  return true;
	}

      } // End of namespace internals.


      template <class I1, class I2, class B>
      fast_intersector<I1, I2, B>::
      fast_intersector(const I1& first1, const I1& last1,
		       const I2& first2, const I2& last2):
	first_ (first1),
	last_ (last1)
# ifndef DONT_USE_HASH
	, arrivals_ (hash_bucket_count)
# endif
      {
	for (I2 i = first2; i != last2; ++i)
	  {
	    typedef typename arrivals_type::iterator	iterator;

	    std::pair<iterator, bool> r = arrivals_.insert(*i);

	    if (not r.second)
	      std::cerr << "WARNING: Unique frame " << r.first->frame_id()
			<< " (" << r.first->microseconds() << " us) is not "
			   "really unique. Frame " << i->frame_id() << " ("
			<< i->microseconds() << " us) is a duplicate. "
		           "Ignoring the last frame." << std::endl;
	  }
      }

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_FAST_INTERSECTOR_HXX_
