/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_SYNCHRONIZER_HXX_
# define WIFI_FRAME_FILTER_SYNCHRONIZER_HXX_

# include "linear_regression_synchronizer.hh"

# include <trace-tools/tool/linear_regression.hh>
# include <trace-tools/tool/window.hh>
# include <trace-tools/wifi/frame/filter/intersector.hh>

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      template <class OriginalFrameType, class ImplType>
      synchronized_frame<OriginalFrameType, ImplType>::
      synchronized_frame(const OriginalFrameType&	orig,
			 const coefs_type&		coefs):
	OriginalFrameType (orig),
	coefs (coefs)
      {
      }

      namespace internals
      {

	template <class I, class B1, class B2>
	lr_sync_iterator<I, B1, B2>::
	lr_sync_iterator(const iterable_type& iterable, bool end):
	  super_type (),
	  iterable_ (&iterable),
	  next_ (end ? iterable.last_ : iterable.first_),
	  next_wpos_ (0)
	{
	  if (not end)
	    increment();
	}

	template <class I, class B1, class B2>
	bool
	lr_sync_iterator<I, B1, B2>::equal(const exact_type& rhs) const
	{
	  if (not rhs.value())
	    return not this->value();
	  return next_wpos_ == rhs.next_wpos_ and next_ == rhs.next_;
	}

	template <class I, class B1, class B2>
	void
	lr_sync_iterator<I, B1, B2>::increment()
	{
	  typedef typename value_type::impl_type	impl_type;

	  typedef
	    typename tool::lr::pair_with_microseconds<impl_type>
	    adapter_type;

	  const I&	last = iterable_->last_;
	  const size_t	window_size = next_->size();

	  if (next_wpos_ == window_size and next_ == last)
	    {
	      this->value() = boost::none_t ();
	      return;
	    }

	  this->value() =
	    value_type (next_[next_wpos_],
			tool::linear_regression<impl_type,
						adapter_type> (next_->begin(),
							       next_->end()));

	  if (next_wpos_ < window_size / 2)
	    ++next_wpos_;
	  else
	    {
	      if (next_ != last)
		++next_;
	      if (next_ == last)
		++next_wpos_;
	    }
	}

      } // End of namespace wifi::frame::filter::internals.

      template <class I, class B>
      linear_regression_synchronizer<I, B>::
      linear_regression_synchronizer(const I& first, const I& last):
	first_ (first),
	last_ (last)
      {
      }

      namespace internals
      {

	template <class F>
	struct provide_lr_synchronizer
	{
	  enum { window_size = 3 };

	  provide_lr_synchronizer(F& func): func_ (func)
	  {
	  }

	  template <class Intersector>
	  void
	  operator () (const Intersector& i)
	  {
	    typedef typename Intersector::const_iterator	i_iterator;
	    typedef tool::window_maker<i_iterator, window_size>	window_maker;
	    typedef typename window_maker::const_iterator	w_iterator;
	    typedef linear_regression_synchronizer<w_iterator>	synchronizer;

	    window_maker	w (i.begin(), i.end());
	    synchronizer	s (w.begin(), w.end());

	    func_(s);
	  }

	private:
	  F&	func_;
	};

      } // End of namespace wifi::frame::filter::internals.

      template <class U, class HT, template <class, class, class> class Int,
		class I1, class I2, class F>
      void
      provide_lr_synchronizer(const I1& first1, const I1& last1,
			      const I2& first2, const I2& last2,
			      addr_mapping&		mapping,
			      F&			func,
			      bool			filter_prism,
			      tool::endian::endianness	phy_end)
      {
	internals::provide_lr_synchronizer<F> func2 (func);

	provide_intersector<U, HT, Int>(first1, last1,
					first2, last2,
					mapping, func2, filter_prism, phy_end);
      }


    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_SYNCHRONIZER_HXX_
