/*
 * trace-tools - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_SIMPLE_MERGE_HXX_
# define WIFI_FRAME_FILTER_SIMPLE_MERGE_HXX_

# include "simple_merge.hh"

# include <trace-tools/prism_header.hh>
# include <trace-tools/wifi/frame/filter/time_adjuster.hh>

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      namespace internals
      {

	template <class HT, class I1, class I2, class B1, class B2>
	simple_merge_iterator<HT, I1, I2, B1, B2>::
	simple_merge_iterator(const iterable_type& i, bool end):
	  iterable_ (&i),
	  next1_ (end ? i.last1_ : i.first1_),
	  next2_ (end ? i.last2_ : i.first2_),
	  current_ (first_packet())
	{
	}

	template <class HT, class I1, class I2, class B1, class B2>
	bool
	simple_merge_iterator<HT, I1, I2, B1, B2>::
	equal(const simple_merge_iterator& rhs) const
	{
	  return next1_ == rhs.next1_ and next2_ == rhs.next2_;
	}

	template <class HT, class I1, class I2, class B1, class B2>
	void
	simple_merge_iterator<HT, I1, I2, B1, B2>::increment()
	{
	  assert(current_ == 0 or current_ == 1);

	  if (next1_ == iterable_->last1_)
	    ++next2_;
	  else if (next2_ == iterable_->last2_)
	    ++next1_;
	  else if (internals::equal<HT>(*next1_, *next2_, iterable_->phy_end_))
	    {
	      ++next1_;
	      ++next2_;
	    }
	  else
	    {
	      if (current_)
		++next2_;
	      else
		++next1_;
	    }
	  current_ = first_packet();
	}

	template <class HT, class I1, class I2, class B1, class B2>
	const typename simple_merge_iterator<HT, I1, I2, B1, B2>::value_type&
	simple_merge_iterator<HT, I1, I2, B1, B2>::get() const
	{
	  assert(current_ == 0 or current_ == 1);

	  if (current_)
	    return *next2_;
	  else
	    return *next1_;
	}

	template <class HT, class I1, class I2, class B1, class B2>
	typename simple_merge_iterator<HT, I1, I2, B1, B2>::value_type&
	simple_merge_iterator<HT, I1, I2, B1, B2>::get()
	{
	  assert(current_ == 0 or current_ == 1);

	  if (current_)
	    return next2_.get();
	  else
	    return next1_.get();
	}

	template <class HT, class I1, class I2, class B1, class B2>
	const typename simple_merge_iterator<HT, I1, I2, B1, B2>::value_type*
	simple_merge_iterator<HT, I1, I2, B1, B2>::get_ptr() const
	{
	  assert(current_ == 0 or current_ == 1);

	  if (current_)
	    return next2_.get_ptr();
	  else
	    return next1_.get_ptr();
	}

	template <class HT, class I1, class I2, class B1, class B2>
	typename simple_merge_iterator<HT, I1, I2, B1, B2>::value_type*
	simple_merge_iterator<HT, I1, I2, B1, B2>::get_ptr()
	{
	  assert(current_ == 0 or current_ == 1);

	  if (current_)
	    return next2_.get_ptr();
	  else
	    return next1_.get_ptr();
	}

	template <class HT, class I1, class I2, class B1, class B2>
	int
	simple_merge_iterator<HT, I1, I2, B1, B2>::first_packet() const
	{
	  const bool first_at_end	= next1_ == iterable_->last1_;
	  const bool second_at_end	= next2_ == iterable_->last2_;

	  if (first_at_end)
	    return second_at_end ? -1 : 1;
	  if (second_at_end)
	    return 0;

	  return next1_->microseconds() < next2_->microseconds() ? 0 : 1;
	}

	template <>
	inline
	bool
	equal<prism::header>(const pcapxx::frame_descriptor&	lhs,
			     const pcapxx::frame_descriptor&	rhs,
			     tool::endian::endianness		phy_end)
	{
	  using tool::endian::need_swap;

	  typedef prism::header		bt;
	  typedef pcapxx::pkthdr	pt;

	  const pt* const lp = lhs.pcap_header().get();
	  const bool	  ls = need_swap(phy_end, lp->swapped);
	  const bt* const lb = reinterpret_cast<const bt*> (lhs.bytes().get());

	  const pt* const rp = rhs.pcap_header().get();
	  const bool	  rs = need_swap(phy_end, rp->swapped);
	  const bt* const rb = reinterpret_cast<const bt*> (rhs.bytes().get());

	  return eq_time_and_80211(lb, lp->caplen, ls,
				   rb, rp->caplen, rs);
	}

      } // End of namespace internals.


      template <class HeaderType,
		class InputIterator1,
		class InputIterator2,
		class Bottom>
      simple_merge<HeaderType, InputIterator1, InputIterator2, Bottom>::
      simple_merge(const InputIterator1& first1, const InputIterator1& last1,
		   const InputIterator2& first2, const InputIterator2& last2,
		   tool::endian::endianness	phy_end):
	phy_end_ (phy_end),
	first1_ (first1),
	last1_ (last1),
	first2_ (first2),
	last2_ (last2)
      {
      }

      namespace internals
      {

	template <class HT, class I, class F>
	struct provide_simple_merge
	{
	  provide_simple_merge(const I&			first1,
			       const I&			last1,
			       F&			func,
			       bool			filter_prism,
			       tool::endian::endianness	phy_end):
	    first1_ (first1),
	    last1_ (last1),
	    func_ (func),
	    filter_prism_ (filter_prism),
	    phy_end_ (phy_end)
	  {
	  }

	  template <class TimeAdjuster>
	  void
	  operator () (const TimeAdjuster& a)
	  {
	    if (filter_prism_)
	      {
		typedef non_noisy_prism<I>			nnp;
		typedef typename nnp::const_iterator		n_iterator;
		typedef microseconds_stamper<n_iterator, HT>	us_stamper;
		typedef typename us_stamper::const_iterator	u_iterator;
		typedef typename TimeAdjuster::const_iterator	a_iterator;
		typedef simple_merge<HT, u_iterator, a_iterator> merger;

		nnp		n (first1_, last1_);
		us_stamper	s (n.begin(), n.end(), phy_end_);
		merger		m (s.begin(), s.end(),
				   a.begin(), a.end(), phy_end_);

		func_(m);
	      }
	    else
	      {
		typedef microseconds_stamper<I, HT>		us_stamper;
		typedef typename us_stamper::const_iterator	u_iterator;
		typedef typename TimeAdjuster::const_iterator	a_iterator;
		typedef simple_merge<HT, u_iterator, a_iterator> merger;

		us_stamper	s (first1_, last1_, phy_end_);
		merger		m (s.begin(), s.end(),
				   a.begin(), a.end(), phy_end_);

		func_(m);
	      }
	  }

	private:
	  const I&			first1_;
	  const I&			last1_;
	  F&				func_;
	  bool				filter_prism_;
	  tool::endian::endianness	phy_end_;
	};

      }

      template <class U, class HT, template <class, class, class> class Int,
		class I1, class I2, class F>
      void
      provide_simple_merge(const I1& first1, const I1& last1,
			   const I2& first2, const I2& last2,
			   addr_mapping&		mapping,
			   F&				func,
			   bool				filter_prism,
			   tool::endian::endianness	phy_end)
      {
	internals::provide_simple_merge<HT, I1, F> func2 (first1, last1,
							  func,
							  filter_prism,
							  phy_end);

	provide_time_adjuster<U, HT, Int>(first1, last1,
					  first2, last2,
					  mapping, func2,
					  filter_prism, phy_end);
      }

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_SIMPLE_MERGE_HXX_
